package Net::Server;

##################################
# Author: Paul T Seamons
# $Id: Server.pm,v 1.57 2001/02/26 14:23:20 rhandom Exp $
# Documentation follows at bottom
##################################

use strict;
use vars qw($VERSION);
use Socket qw( inet_ntoa unpack_sockaddr_in AF_INET );
use IO::Socket ();


$VERSION = '0.4.1';

### program flow
sub run {

  ### pass package or object
  my $self = ref($_[0]) ? shift() : (bless {}, shift());

  $self->pre_configure_hook;  # user customizable hook

  $self->configure(@_);       # allow for reading of commandline,
                              # program, and configuration file parameters

  $self->post_configure;      # verification of passed parameters

  $self->post_configure_hook; # user customizable hook

  $self->pre_bind;            # finalize ports to be bound

  $self->bind;                # connect to port(s)
                              # setup selection handle for multi port

  $self->post_bind_hook;      # user customizable hook

  $self->post_bind;           # allow for chrooting,
                              # becoming a different user and group

  $self->pre_loop_hook;      # user customizable hook

  $self->loop;                # repeat accept/process cycle

  ### routines inside a standard $self->loop
  # $self->accept             # wait for client connection
  # $self->run_client_connection # process client
  # $self->done               # indicate if connection is done

  $self->pre_server_close_hook;  # user customizable hook

  $self->server_close;        # close the server and release the port

  exit;
}

### standard connection flow
sub run_client_connection {
  my $self = shift;

  $self->post_accept;         # prepare client for processing

  $self->get_client_info;     # determines information about peer and local

  $self->post_accept_hook;    # user customizable hook

  if( $self->allow_deny             # do allow/deny check on client info
      && $self->allow_deny_hook ){  # user customizable hook
    
    $self->process_request;   # This is where the core functionality
                              # of a Net::Server should be.  This is the
                              # only method necessary to override.
  }else{

    $self->request_denied_hook;     # user customizable hook

  }

  $self->post_process_request_hook; # user customizable hook

  $self->post_process_request;      # clean up client connection, etc

}


###----------------------------------------------------------###


### any pre-initialization stuff
sub pre_configure_hook {}


### set up the object a little bit better
sub configure {
  my $self = shift;

  ### do command line
  $self->process_args( \@ARGV ) if defined @ARGV;

  ### do startup file args
  $self->process_args( \@_ ) if defined @ARGV;

  ### do a config file
  if( defined $self->{conf_file} ){
    $self->process_conf( $self->{conf_file} );
  }

}


### make sure it has been configured properly
sub post_configure {
  my $self = shift;

  ### set the log level
  if( !defined $self->{log_level} || $self->{log_level} !~ /^\d+$/ ){
    $self->{log_level} = 1;
  }
  $self->{log_level} = 5 if $self->{log_level} > 5;

  ### open a log file (any children will share the open connection)
  if( defined $self->{log_file} ){
    die "Unsecure filename \"$self->{log_file}\""
      unless $self->{log_file} =~ m|^([\w\.\-/]+)$|;
    $self->{log_file} = $1;
    open(_SERVER_LOG, ">>$self->{log_file}")
      or die "Couldn't open log file \"$self->{log_file}\" [$!].";
  }

  ### background the process
  if( defined $self->{background} ){
    my $pid = fork;
    if( not defined $pid ){ $self->fatal("Couldn't fork [$!]"); }
    exit if $pid;
    $self->log(2,"Process backgrounded");
  }

  ### allow for a pid file (must be done after backgrounding)
  if( defined $self->{pid_file} ){
    unless( $self->{pid_file} =~ m|^([\w\.\-/]+)$| ){
      $self->fatal("Unsecure filename \"$self->{pid_file}\"");
    }
    $self->{pid_file} = $1;
    if( -e $self->{pid_file} ){
      $self->fatal("Pid file \"$self->{pid_file}\" all ready exists.");
    }elsif( open(PID, ">$self->{pid_file}") ){
      print PID $$;
      close PID;
    }else{
      $self->fatal("Couldn't open pid file \"$self->{pid_file}\" [$!].");
    }
  }

  ### make sure that allow and deny look like array refs
  $self->{allow} = [] unless defined($self->{allow}) && ref($self->{allow});
  $self->{deny}  = [] unless defined($self->{deny})  && ref($self->{deny} );

}


### user customizable hook
sub post_configure_hook {}


### make sure we have good port parameters
sub pre_bind {
  my $self = shift;

  $self->log(1,$self->log_time ." ". ref($self) ." starting!!!");

  ### set a default port
  if( ! defined( $self->{port} )
      || ! ref( $self->{port} )
      || ! @{ $self->{port} } ){
    $self->log(1,"Port Not Defined.  Defaulting to '20203'\n");
    $self->{port}  = [ 20203 ];
  }

  ### loop through the passed ports
  ### set up parallel arrays of hosts, ports, and protos
  foreach (@{ $self->{port} }){
    if( m|^([\w\.\-\/]+):(\w+)/(\w+)$| ){
      $_ = "$1:$2:$3";
    }elsif( /^([\w\.\-]+):(\w+)$/ ){
      my $proto = defined $self->{proto} ? $self->{proto} : 'tcp';
      $_ = "$1:$2:$proto";
    }elsif( /^(\w+)$/ ){
      my $host  = defined $self->{host}  ? $self->{host}  : 'localhost';
      my $proto = defined $self->{proto} ? $self->{proto} : 'tcp';
      $_ = "$host:$1:$proto";
    }else{
      $self->fatal("Undeterminate port \"$_\"");
    }
  }
  if( @{ $self->{port} } < 1 ){
    $self->fatal("No valid ports found");
  }

}


### bind to the port (This should serve all but INET)
sub bind {
  my $self = shift;

  ### connect to any ports
  foreach (my($i)=0 ; $i<=$#{ $self->{port} } ; $i++){
    my ($host,$port,$proto) = split(/:/, $self->{port}->[$i]);
    $self->log(1,"Binding to $proto port $port on host $host\n");
    $self->{sock}->[$i] =
      IO::Socket::INET->new
        (LocalAddr => $host,
         LocalPort => $port,
         Proto     => $proto,
         Listen    => 5,
         Reuse     => 1,
         )
          or $self->fatal("Can't connect to $proto port $port on $host [$!]");

    $self->fatal("Back sock [$i]!")
      unless $self->{sock}->[$i];
  }

  ### if more than one port, we'll need to select on it
  if( @{ $self->{sock} } > 1 ){
    $self->{multi_port} = 1;
    require "IO/Select.pm";
    $self->{select} = IO::Select->new();
    foreach ( @{ $self->{sock} } ){
      $self->{select}->add( $_ );
    }
  }else{
    $self->{multi_port} = undef;
    $self->{select}     = undef;
  }

}


### user customizable hook
sub post_bind_hook {}


### secure the process and background it
sub post_bind {

  my $self = shift;

  ### perform the chroot operation
  if( defined $self->{chroot} ){
    if( ! -d $self->{chroot} ){
      $self->fatal("Specified chroot \"$self->{chroot}\" doesn't exist.\n");
    }else{
      $self->log(2,"Chrooting to $self->{chroot}\n");
      chroot( $self->{chroot} )
        or $self->fatal("Couldn't chroot to \"$self->{chroot}\"");
    }
  }
  
  ### become another group
  if( !defined $self->{group} ){
    $self->log(1,"Group Not Defined.  Defaulting to 'nobody'\n");
    $self->{group}  = 'nobody';
  }
  $self->log(2,"Setting group to $self->{group}\n");
  $self->{group} = getgrnam( $self->{group} )
    if $self->{group} =~ /\D/;
  $( = $) = $self->{group};

  ### become another user
  if( !defined $self->{user} ){
    $self->log(1,"User Not Defined.  Defaulting to 'nobody'\n");
    $self->{user}  = 'nobody';
  }
  $self->log(2,"Setting user to $self->{user}\n");
  $self->{user}  = getpwnam( $self->{user} )
    if $self->{user}  =~ /\D/;
  $< = $> = $self->{user};

  $self->{requests} = 0;

}


### user customizable hook
sub pre_loop_hook {}


### receive requests
sub loop {
  my $self = shift;

  while( $self->accept() ){

    $self->run_client_connection;

    last if $self->done;

  }
}


### wait for the connection
sub accept {
  my $self = shift;
  my $retries  = 30;
  my $sock = undef;

  ### try awhile to get a defined client handle
  ### normally a good handle should occur every time
  while( $retries-- ){

    ### with more than one port, use select to get the next one
    if( defined $self->{multi_port} ){

      ### anything server type specific
      $sock = $self->accept_multi_port;
      
    ### single port is bound - just accept
    }else{

      $sock = $self->{sock}->[0];

    }

    ### make sure we got a good sock
    if( not defined $sock ){
      $self->fatal("Recieved a bad sock!");
    }

    $self->{client} = $sock->accept();
    
    ### success
    return 1 if defined $self->{client};

    ### try again in a second
    sleep(1);

  }

  return undef;
}


### server specific hook for multi port applications
### this actually applies to all but INET
sub accept_multi_port {
  my $self = shift;

  if( not exists $self->{select} ){
    $self->fatal("No select property during multi_port execution.");
  }

  ### this will block until a client arrives
  my @waiting = $self->{select}->can_read();

  ### choose a socket
  return $waiting[ rand(@waiting) ];

}


### this occurs after the request has been processed
### this is server type specific (actually applies to all by INET)
sub post_accept {
  my $self = shift;

  ### keep track of the requests
  $self->{requests} ++;

  ### duplicate some handles and flush them
  ### maybe we should save these somewhere - maybe not
  *STDIN  = \*{ $self->{client} };
  *STDOUT = \*{ $self->{client} };
  STDIN->autoflush(1);
  STDOUT->autoflush(1);
  select(STDOUT);

}


### read information about the client connection
sub get_client_info {
  my $self = shift;

  ### read information about this connection
  my $sockname = getsockname( STDIN );
  if( $sockname ){
    ($self->{sockport}, $self->{sockaddr})
      = unpack_sockaddr_in( $sockname );
    $self->{sockaddr} = inet_ntoa( $self->{sockaddr} );

  }else{
    ### does this only happen from command line?
    $self->{sockaddr} = '0.0.0.0';
    $self->{sockhost} = 'inet.test';
    $self->{sockport} = 0;
  }
  
  ### try to get some info about the remote host
  my $peername = getpeername( STDIN );
  if( $peername ){
    ($self->{peerport}, $self->{peeraddr})
      = unpack_sockaddr_in( $peername );
    $self->{peeraddr} = inet_ntoa( $self->{peeraddr} );

    if( defined $self->{reverse_lookups} ){
      $self->{peerhost} = gethostbyaddr( $self->{peeraddr}, AF_INET );
    }
    $self->{peerhost} = '' unless defined $self->{peerhost};

  }else{
    ### does this only happen from command line?
    $self->{peeraddr} = '0.0.0.0';
    $self->{peerhost} = 'inet.test';
    $self->{peerport} = 0;
  }

  $self->log(3,$self->log_time
             ." CONNECT peer: \"$self->{peeraddr}:$self->{peerport}\""
             ." local: \"$self->{sockaddr}:$self->{sockport}\"\n");

}

### user customizable hook
sub post_accept_hook {}


### perform basic allow/deny service
sub allow_deny {
  my $self = shift;

  ### if not allow or deny parameters are set, allow all
  return 1 unless @{ $self->{allow} } || @{ $self->{deny} };

  ### if the addr or host matches a deny, reject it immediately
  foreach ( @{ $self->{deny} } ){
    return 0 if $self->{peerhost} =~ /^$_$/ && defined($self->{reverse_lookups});
    return 0 if $self->{peeraddr} =~ /^$_$/;
  }

  ### if the addr or host isn't blocked yet, allow it if it is allowed
  foreach ( @{ $self->{allow} } ){
    return 1 if $self->{peerhost} =~ /^$_$/ && defined($self->{reverse_lookups});
    return 1 if $self->{peeraddr} =~ /^$_$/;
  }

  return 0;
}


### user customizable hook
### if this hook returns 1 the request is processed
### if this hook returns 0 the request is denied
sub allow_deny_hook { 1 }


### user customizable hook
sub request_denied_hook {}


### this is the main method to override
### this is where most of the work will occur
### A sample server is shown below.
sub process_request {
  my $self   = shift;

  print "Welcome to \"".ref($self)."\" ($$)\n";

  ### eval block needed to prevent DoS by using timeout
  eval {

    local $SIG{ALRM} = sub { die "Timed Out!\n" };
    my $timeout = 30; # give the user 30 seconds to type a line
    my $previous_alarm = alarm($timeout);

    while( <STDIN> ){
      
      s/\r?\n$//;
      
      print ref($self),":$$: You said \"$_\"\r\n";
      $self->log(5,$_); # very verbose log
      
      if( /dump/ ){
        require "Data/Dumper.pm";
        print Data::Dumper::Dumper( $self );
      }
      
      if( /quit/ ){ last }
      
      if( /exit/ ){ $self->server_close }
      
      alarm($timeout);
    }
    alarm($previous_alarm);

  };

  if( $@=~/timed out/i ){
    print STDOUT "Timed Out.\r\n";
    return;
  }

}


### user customizable hook
sub post_process_request_hook {}


### this is server type specific functions after the process
sub post_process_request {
  my $self = shift;

  ### close the socket handle
  close($self->{client});

}


### determine if I am done with a request
### in the base type, we are never done until a SIG occurs
sub done { 0 }


### user customizable hook
sub pre_server_close_hook {}


### this happens when the server reaches the end
sub server_close{
  my $self = shift;
  if( defined $self->{pid_file} && -e $self->{pid_file} ){
    unlink $self->{pid_file};
  }
  exit;
}


### what to do when all else fails
sub fatal {
  my $self = shift;
  my $error = shift;
  my ($pack,$file,$line) = caller;

  $self->log(0, $self->log_time ." ". $error
             ."\n  at line $line in file $file");

  $self->server_close;
}


### record output
sub log {
  my $self  = shift;
  my $level = shift;
  return unless $self->{log_level};
  return unless $level <= $self->{log_level};

  warn "$_[0]\n" if ! $level and defined $self->{log_file};

  $self->write_to_log(@_);
}


### standard log routine, this could very easily be
### overridden with a syslog call
sub write_to_log {
  my $self = shift;

  local $_ = shift || '';
  chomp;
  s/([^\n\ -\~])/sprintf("%%%02X",ord($1))/eg;

  if( defined $self->{log_file} ){
    print _SERVER_LOG $_, "\n";
  }else{
    my $old = select(STDERR);
    print $_, "\n";
    select($old);
  }

}


### default time format
sub log_time {
  my ($sec,$min,$hour,$day,$mon,$year) = localtime;
  return sprintf("%04d/%02d/%02d-%02d:%02d:%02d",
                 $year+1900, $mon+1, $day, $hour, $min, $sec);
}


### set up default structure
sub options {
  my $self = shift;
  my $ref  = shift;

  foreach ( qw(port allow deny) ){
    $self->{$_} = [] unless exists $self->{$_};
    $ref->{$_} = $self->{$_};
  }

  foreach ( qw(user group chroot log_level
               log_file pid_file background
               host proto) ){
    $self->{$_} = undef unless exists $self->{$_};
    $ref->{$_} = \$self->{$_};
  }

}


### routine for parsing commandline, module, and conf file
### possibly should use Getopt::Long but this
### method has the benefit of leaving unused arguments in @ARGV
sub process_args {
  my $self = shift;
  my $ref  = shift;
  my $template = {};
  $self->options( $template );

  foreach (my $i=0 ; $i < @$ref ; $i++ ){
    local $_ = $ref->[$i];

    if( /^(?:--)?(\w+)([=\ ](\S+))?$/ && exists $template->{$1} ){
      my ($key,$val) = ($1,$3);
      splice( @$ref, $i, 1 );
      $val = splice( @$ref, $i, 1 ) if not defined $val;
      $i--;
      $val =~ s/%([A-F0-9])/chr(hex($1))/eig;

      if( ref $template->{$key} eq 'ARRAY' ){
        push( @{ $template->{$key} }, $val );
      }else{
        $ { $template->{$key} } = $val;
      }
    }
  }

}


### routine for loading conf file parameters
sub process_conf {
  my $self = shift;
  my $file = shift;
  my @args = ();

  $file = ($file =~ m|^([\w\.\-/]+)$|)
    ? $1 : $self->fatal("Unsecure filename \"$file\"");

  if( not open(_CONF,"<$file") ){
    $self->fatal("Couldn't open conf \"$file\" [$!]");
  }

  while(<_CONF>){
    push( @args, "$1=$2") if m/^\s*((?:--)?\w+)[=:]?\s*(\S+)/;
  }

  close(_CONF);

  $self->process_args( \@args );
}


1;

__END__

=head1 NAME

Net::Server - Extensible, general Perl server engine

=head1 SYNOPSIS

  use Net::Server; @ISA = qw(Net::Server);

  sub process_request { #...code...  }

  Net::Server->run();

=head1 FEATURES

 * Single Server Mode
 * Inetd Server Mode
 * Preforking Mode
 * Forking Mode
 * Multi port accepts on Single, Preforking, and Forking modes
 * User customizable hooks
 * Chroot ability after bind
 * Change of user and group after bind
 * Taint clean
 * Written in Perl
 * Protection against buffer overflow
 * Clean process flow
 * Extensibility

=head1 DESCRIPTION

C<Net::Server> is an extensible, generic Perl server engine.
C<Net::Server> combines the good properties from
C<Net::Daemon> (0.34), C<NetServer::Generic> (1.03), and
C<Net::FTPServer> (1.0), and also from various concepts in
the Apache Webserver.

C<Net::Server> attempts to be a generic server as in
C<Net::Daemon> and C<NetServer::Generic>.  It includes with
it the ability to run as an inetd process
(C<Net::Server::INET>), a single connection server
(C<Net::Server> or C<Net::Server::Single>), a forking server
(C<Net::Server::Fork>), or as a preforking server
(C<Net::Server::PreFork>).  In all but the inetd type, the
server provides the ability to connect to one or to multiple
server ports.

C<Net::Server> uses ideologies of C<Net::FTPServer> in order
to provide extensibility.  The additional server types are
made possible via "personalities" or sub classes of the
C<Net::Server>.  By moving the multiple types of servers out of
the main C<Net::Server> class, the C<Net::Server> concept is
easily extended to other types (in the near future, we would
like to add a "Thread" personality).  C<Net::Server> borrows
several concepts from the Apache Webserver.  C<Net::Server>
uses "hooks" to allow custom servers such as SMTP, HTTP,
POP3, etc. to be layered over the base C<Net::Server> class.
In addition the C<Net::Server::PreFork> class borrows concepts
of min_start_servers, max_servers, and min_waiting servers.
C<Net::Server::PreFork> also uses the concept of an flock
serialized accept when accepting on multiple ports.

=head1 SAMPLE

The following is a very simple server.  The main
functionality occurs in the process_request method call as
shown below.  (Other examples of using C<Net::Server> can, or
will, be included with this distribution)

  #!/usr/bin/perl -w -T
  #--------------- file test.pl ---------------

  MyPackage->run();
  exit;


  package MyPackage;

  use strict;
  use vars qw(@ISA);
  use Net::Server::PreFork; # any personality will do

  @ISA = qw(Net::Server::PreFork);

  sub process_request {
    my $self = shift;
    while (<STDIN>) {
      chomp;
      print "You said '$_'\r\n";
    }
  }

  1;

  #--------------- file test.pl ---------------

=head1 COMMAND LINE ARGUMENTS

Command line arguments consist of key value pairs.  These
pairs follow the POSIX fashion of C<--key value> or
C<--key=value>.  Key/value pairs used by the server are
removed by the configuration process so that server layers
on top of C<Net::Server> can pass and read their own
parameters.  Currently, Getopt::Long is not used.

=head1 CONFIGURATION FILE

C<Net::Server> allows for the use of a configuration file to
read in server parameters.  The format of this conf file is
simple key value pairs.  Comments and white space are
ignored.

  #-------------- file test.conf --------------
  key value

  #-------------- file test.conf --------------

=head1 PROCESS FLOW

The process flow is written in an open, easy to
override, easy to hook, fashion.  The basic flow is
shown below.

  $self->pre_configure_hook;

  $self->configure(@_);

  $self->post_configure;

  $self->post_configure_hook;

  $self->pre_bind;

  $self->bind;

  $self->post_bind_hook;

  $self->post_bind;

  $self->pre_loop_hook;

  $self->loop;

  ### routines inside a standard $self->loop
  # $self->accept;
  # $self->run_client_connection;
  # $self->done;

  $self->pre_server_close_hook;

  $self->server_close;

The server then exits.

During the client processing phase
(C<$self-E<gt>run_client_connection>), the following
represents the program flow:

  $self->post_accept;

  $self->get_client_info;

  $self->post_accept_hook;

  if( $self->allow_deny

      && $self->allow_deny_hook ){
    
    $self->process_request;

  }else{

    $self->request_denied_hook;

  }

  $self->post_process_request_hook;

  $self->post_process_request;

The process then loops and waits for the next
connection.  For a more in depth discussion, please
read the code.


=head1 HOOKS

C<Net::Server> provides a number of "hooks" allowing for
servers layered on top of C<Net::Server> to respond at
different levels of execution.

=over 4

=item C<$self-E<gt>pre_configure_hook()>

This hook takes place immediately after the C<-E<gt>run()>
method is called.  This hook allows for setting up the
object before any configuration parameters are read in.

=item C<$self-E<gt>post_configure_hook()>

This hook occurs just after the reading of configuration
parameters and initiation of logging and pid_file creation.
It also occurs before the C<-E<gt>pre_bind()> and
C<-E<gt>bind()> methods are called.  This hook allows for
verifying configuration parameters.

=item C<$self-E<gt>post_bind_hook()>

This hook occurs just after the bind process and just before
any chrooting, change of user, or change of group occurs.
At this point the process will still be running as the user
who started the server.

=item C<$self-E<gt>pre_loop_hook()>

This hook occurs after chroot, change of user, and change of
group has occured.  It allows for preparation before looping
begins.

=item C<$self-E<gt>post_accept_hook()>

This hook occurs after a client has connected to the server.
At this point STDIN and STDOUT are mapped to the client
socket.  This hook occurs before the processing of the
request.

=item C<$self-E<gt>allow_deny_hook()>

This hook allows for the checking of ip and host information
beyond the C<$self-E<gt>allow_deny()> routine.  If this hook
returns 1, the client request will be processed,
otherwise, the request will be denied processing.

=item C<$self-E<gt>request_denied_hook()>

This hook occurs if either the C<$self-E<gt>allow_deny()> or
C<$self-E<gt>allow_deny_hook()> have taken place.

=item C<$self-E<gt>post_process_request_hook()>

This hook occurs after the processing of the request, but
before the client connection has been closed.

=item C<$self-E<gt>pre_server_close_hook()>

This hook occurs before the server begins shutting down.

=back

=head1 TO DO

There are several tasks to perform before the alpha label
can be removed from this software:

=over 4

=item Use It

The best way to further the status of this project is to use
it.  There are immediate plans to use this as a base class
in implementing some mail servers and banner servers on a
high hit site.

=item Thread Personality

Some servers offer a threaded server.  Create
C<Net::Server::Thread> as a new personality.

=item Other Personalities

Explore any other personalities

=item Sig Handling

Solidify which signals are handled by base class.  Possibly
catch more that are ignored currently.

=item C<HUP>

Allow for a clean hup allowing for re-exec and re-read of
configuration files.  This includes multiport mode.

=item Net::HTTPServer, etc

Create various types of servers.  Possibly, port exising
servers to user Net::Server as a base layer.

=item More documentation

Show more examples and explain process flow more.

=item Better Tests

Do better tests during "make test"

=back

=head1 FILES

  The following files are installed as part of this
  distribution.

  Net/Server.pm
  Net/Server/Fork.pm
  Net/Server/INET.pm
  Net/Server/MultiType.pm
  Net/Server/PreFork.pm
  Net/Server/Single.pm

=head1 AUTHOR

Paul T. Seamons paul@seamons.com

=head1 SEE ALSO

Please see also
L<Net::Server::Fork>,
L<Net::Server::INET>,
L<Net::Server::PreFork>,
L<Net::Server::MultiType>,
L<Net::Server::Single>

=cut
