package Net::Server::PreFork;

use strict;
use vars qw($VERSION @ISA $LOCK_EX $LOCK_UN);
use subs qw(tmpnam);
use Net::Server;

$VERSION = 0.3;

### fall back to parent methods
@ISA = qw(Net::Server);

### there is something to be said for hard coding numbers,
### this could be done by "use Fcntl qw(LOCK_EX LOCK_UN);"
### however, that adds 70k to each process (24k unshared)
*LOCK_EX = \2;
*LOCK_UN = \8;

### override-able options for this package
sub options {
  my $self = shift;
  my $ref  = shift;

  $self->SUPER::options($ref);

  foreach ( qw(min_servers max_servers spare_servers max_requests
               check_for_dead check_for_waiting) ){
    $self->{$_} = undef unless exists $self->{$_};
    $ref->{$_} = \$self->{$_};
  }

}

### make sure some defaults are set
sub post_configure {
  my $self = shift;

  ### let the parent do the rest
  ### must do this first so that ppid reflects backgrounded process
  $self->SUPER::post_configure;

  ### some default values to check for
  my $d = {min_servers   => 5,      # min num of servers to always have running
           max_servers   => 10,     # max num of servers to run
           spare_servers => 1,      # num of extra servers to have lying around
           max_requests  => 1000,   # num of requests for each child to handle
           check_for_dead    => 30, # how often to see if children are alive
           check_for_waiting => 20, # how often to see if extra children exist
           };
  foreach (keys %$d){
    $self->{$_} = $d->{$_}
    unless defined($self->{$_}) && $self->{$_} =~ /^\d+$/;
  }

  ### I need to know who is the parent
  $self->{ppid} = $$;

}


sub post_bind {
  my $self = shift;

  ### do the parents
  $self->SUPER::post_bind;

  ### I need a lock file if I am multi-port, I need to wait
  ### until I am the final user though so I know I can access it
  if( defined($self->{multi_port}) && not defined($self->{lock_file}) ){
    require "POSIX.pm";
    import POSIX qw(tmpnam);
    $self->{lock_file} = tmpnam;
  }

}

### prepare for connections
sub loop {
  my $self = shift;

  ### get ready for children
  $SIG{CHLD} = 'IGNORE';
  $self->{children} = {};

  ### get ready for child->parent communication
  pipe(_READ,_WRITE);
  $self->{_READ}  = *_READ;
  $self->{_WRITE} = *_WRITE;

  $self->log(2,"Beginning prefork ($self->{min_servers} processes)\n");

  ### start up the children
  $self->run_n_children( $self->{min_servers} );

  ### finish the parent routines
  $self->run_parent;
  
}


### subroutine to start up a specified number of children
sub run_n_children {
  my $self  = shift;
  my $n     = shift;
  my $total = scalar keys %{ $self->{children} };
  
  ### don't start more than we're allowed
  if( $n > $self->{max_servers} - $total ){
    $n = $self->{max_servers} - $total;
  }

  for( 1..$n ){
    my $pid = fork;

    ### trouble
    if( not defined $pid ){
      $self->fatal("Bad fork [$!]");

    ### parent
    }elsif( $pid ){
      $self->{children}->{$pid} = 'waiting';

    ### child
    }else{
      $self->run_child;

    }
  }
}


### child process which will accept on the port
sub run_child {
  my $self = shift;

  $self->log(4,"Child Preforked ($$)\n");
  delete $self->{children};

  ### tell the parent that we are waiting
  *_WRITE = $self->{_WRITE};
  _WRITE->autoflush(1);
  print _WRITE "$$ waiting\n";

  ### let the parent shut me down
  $self->{connected} = 0;
  $self->{SigHUPed}  = 0;
  $SIG{HUP} = sub {
    unless( $self->{connected} ){
      print _WRITE "$$ exiting\n";
      exit;
    }
    $self->{SigHUPed} = 1;
  };

  local *LOCK;

  ### accept connections
  while( $self->accept() ){
    
    ### remove the lock that occured during accept_multi_port
    flock(LOCK,$LOCK_UN) if defined $self->{multi_port};

    $self->{connected} = 1;
    print _WRITE "$$ processing\n";

    $self->run_client_connection;

    last if $self->done;

    $self->{connected} = 0;
    print _WRITE "$$ waiting\n";

  }
  
  print _WRITE "$$ exiting\n";
  exit;

}


### We can only let one process do the selecting at a time
### this override makes sure that nobody else can do it
### while we are.  We do this by opening a lock file
### and getting an exclusive lock - this will block all others
### until we release it (we do this after we have accepted on
### the socket and begin processing the request)
sub accept_multi_port {
  my $self = shift;
  
  open(LOCK,">$self->{lock_file}")
    || $self->fatal("Couldn't open lock file \"$self->{lock_file}\" [$!]");

  flock(LOCK,$LOCK_EX)
    || $self->fatal("Couldn't get lock on file \"$self->{lock_file}\" [$!]");

  return $self->SUPER::accept_multi_port();

}


### is the looping done (non zero value says its done)
sub done {
  my $self = shift;
  return 1 if $self->{requests} >= $self->{max_requests};
  return 1 if $self->{SigHUPed};
}


### now the parent will wait for the kids
sub run_parent {
  my $self=shift;

  $self->log(2,"Parent ready for children\n");

  my $last_checked_for_dead    = time;
  my $last_checked_for_waiting = time;

  *_READ = $self->{_READ};
  _READ->autoflush(1);

  $SIG{INT} = sub { $self->server_close; };

  while(<_READ>){
    my $time = time;

    $self->log(5,"Child_status: $_");

    ### record what the child said
    next unless /^(\d+)\ +(.+)?/;
    my ($pid,$status) = ($1,$2);

    ### record the status
    if( $status eq 'exiting' ){
      delete($self->{children}->{$pid});
    }else{
      $self->{children}->{$pid} = $status;
    }

    ### periodically make sure children are alive
    if( $time - $last_checked_for_dead > $self->{check_for_dead} ){
      $last_checked_for_dead = $time;
      foreach (keys %{ $self->{children} }){
        ### see if the child can be killed
        kill(0,$_) or delete $self->{children}->{$_};
      }
    }

    ### tally the possible types
    my %num = (total=>0, waiting=>0, processing=>0);
    foreach (keys %{ $self->{children} }){
      $num{total} ++;
      $num{ $self->{children}->{$_} } ++;
    }

    ### have all the servers we're going to get
    if( $num{total} >= $self->{max_servers} ){
      # do nothing

    ### need more min_servers
    }elsif( $num{total} < $self->{min_servers} ){
      $self->run_n_children( $self->{min_servers} - $num{total} );

    ### need more spare_servers
    }elsif( $num{waiting} < $self->{spare_servers} ){
      $self->run_n_children( $self->{spare_servers} - $num{waiting} );

    ### need to remove some extra waiting servers
    }elsif( $num{waiting} > $self->{spare_servers} 
            && $num{waiting} + $num{processing} > $self->{min_servers} ){
      if( $time - $last_checked_for_waiting > $self->{check_for_waiting} ){
        $last_checked_for_waiting = $time;
        my $n = $num{waiting} + $num{processing} - $self->{min_servers};
        if( $n > $num{waiting} - $self->{spare_servers} ){
          $n = $num{waiting} - $self->{spare_servers};
        }
        foreach (keys %{ $self->{children} }){
          next unless $self->{children}->{$_} eq 'waiting';
          last unless $n--;
          kill(1,$_);
        }
      }

    }

    ### end of while
  }
}

### routine to shut down the server (and all forked children)
sub server_close {
  my $self = shift;

  ### if a parent, fork off cleanup sub and close
  if( ! defined $self->{ppid} || $self->{ppid} == $$ ){

    $self->log(1,$self->log_time . " Server closing!!!");
    $self->close_children;
    exit;

  ### if a child, signal the parent and close
  ### normally the child shouldn't, but if they do...
  }else{
    kill(2,$self->{ppid});
    exit;
  }

}

### Fork another process (that won't deal with sig chld's) to
### sig int the children and then exit itself
sub close_children {
  my $self = shift;

  if( defined $self->{children} && %{ $self->{children} } ){

    my $pid = fork;
    if( not defined $pid ){
      $self->log(1,"Can't fork to clean children [$!]");
    }
    return if $pid;

    foreach (keys %{ $self->{children} }){
      kill(2,$_);
    }

  }
  
  ### remove extra files
  if( defined $self->{lock_file} && -e $self->{lock_file} ){
    unlink $self->{lock_file};
  }
  if( defined $self->{pid_file} && -e $self->{pid_file} ){
    unlink $self->{pid_file};
  }

  exit;
}


1;
